package MyOAI;

use LWP::UserAgent;
use HTTP::Request;
use XML::Parser;
use Carp;
use strict;
use vars qw(@ISA);

@ISA = qw(LWP::UserAgent);

my $my_oai = '_my_oai';
my $debug = 0;
my $xml_header = "<?xml version=\"1.0\" encoding=\"ISO-8859-1\" standalone=\"yes\"?>";
my $resumptionToken = "resumptionToken";


sub oai_option {
## Sets OAI options, expects syntax key1 => value1, key2 => value2, etc.

    my ($self, %options) = @_;

    foreach (keys %options) {
	$self->{oai_options}->{$_} = $options{$_};
    }
}


sub store_obj {
## Stores object $obj into a referenced hash, and checks if the key is already taken.
    
    my ($storage, $obj) = @_;
    
    if (defined($storage->{$my_oai})) {
	die "$0: Naughty you, overriding existing data member '$my_oai'";
    }
    $storage->{$my_oai} = $obj;
}


sub get_obj {
## Extracts stored object from referenced hash.
    
    my $storage = shift;

    return $storage->{$my_oai};
}


sub set_callback {
## Sets a callback for an element using syntax element => \&my_callback
## At this point, only one element/callback pair can be assigned.

    my ($self, %handles) = @_;
    
    foreach (keys %handles) {
	my $handle = { callback	=> $handles{$_}, seen => 0 };
	$self->{handles}->{$_} = $handle;
    }
}


sub get_callback {
## Returns callback code reference for $element.

    my ($self, $element) = @_;

    return $self->{handles}->{$element}->{callback};
}


sub incomplete {
## Returns the resumptionToken if document is incomplete, otherwise undef.
    
    my $self = shift;
    return $self->{$resumptionToken};
}


sub cursor {
## Returns record offset cursor in case of incomplete document retrieval

    my $self = shift;
    return $self->{cursor};
}


sub completeListSize {
## Returns the total number of records to be returned if incomplete document

    my $self = shift;
    return $self->{completeListSize};
}


sub expirationDate {
## Returns expirationDate of resumptionToken (ehh, at least this is my best guess).

    my $self = shift;
    return $self->{expirationDate};
}


sub get_elements {
## Returns a list of elements with callback.

    my $self = shift;
    return keys %{$self->{handles}};
}


sub harvested {
## Returns number of harvested records

    my $self = shift;
    return $self->{harvested};
}


sub reset_xml {
## Resets XML buffer

    my ($self) = @_;

    $self->{raw_xml} = "";
}


sub store_xml {
## Store a chunk of XML by reference

    my ($self, $xml_ref) = @_;

    $self->{raw_xml} .= $$xml_ref;
}
    

sub get_xml {
## Returns the XML buffer

    my ($self) = @_;

    return $self->{raw_xml};
}


sub oai_request {
## Send an OAI request to the OAI server specified in option url
## The request is specified as arguments of the form key1 => value1, key2 => value2, etc.

    my ($self, @settings) = @_;
    my @old_settings = @{$self->{old_settings} || []};
    my %oai = (@old_settings, @settings);
    my $options = $self->{oai_options};
    my $url = "";
    my $callback = $options->{callback} || \&callback;
    my $chunk_size = $options->{chunksize} || 4096;
    my $parser = new XML::Parser(Handlers => { Start => \&handle_start,
					       End   => \&handle_end,
					       Char  => \&handle_char });
    
    if (!defined($options->{url})) {
	croak "oai_request: You must set OAI option url before calling oai_request";
    }
    if (!defined($oai{verb})) {
	croak "oai_requst: Verb must be specified in OAI request";
    }
    if (defined($self->{$resumptionToken})) {
	$oai{$resumptionToken} = $self->{$resumptionToken};
	delete ($oai{metadataPrefix});
    }
    $self->{$resumptionToken} = undef;
    $self->{expirationDate} = undef;
    $self->{cursor} = undef;
    $self->{completeListSize} = undef;
    $self->{incomplete} = undef;
    $self->{old_settings} = [%oai];
    foreach my $key (keys %oai) {
	if ($url) {
	    $url .= "&";
	} else {
	    $url = "?";
	}
	$url .= "$key=" . $oai{$key};
    }
    $url = $options->{url} . $url;
    #print "URL='$url'\n";
    my $request = new HTTP::Request('GET', $url);
    if (defined($request->{$my_oai})) {
	warn "$0: Naughty you, overriding existing data member '$my_oai'";
    }
    store_obj($request, $self);
    store_obj($parser, $self);
    $self->{chunk_parser} = $parser->parse_start();
    
    return $self->request($request, $callback, $chunk_size);
}


sub get_orig_request {
## Extracts the HTTP request based on the HTTP response.

    my $r = shift;

    while (my $prev = $r->previous()) {
	$r = $prev;
    }
    return $r->request();
}


sub callback {
## This callback is called for each chunk of fetched data

    my ($data, $response, $protocol) = @_;
    my $oai = get_obj(get_orig_request($response));
    my $chunk_parser = $oai->{chunk_parser};

    $chunk_parser->parse_more($data);
}



sub handle_start {
## Expat element start handler

    my ($expat, $element, %attr) = @_;
    my $oai = get_obj($expat);
    my $xml = "<$element";
    my $stack = $oai->{stack};
    
    if ($oai->{triggered}) {
	if ($debug) {
	    print "Pushing '$element'\n";
	}
	push @$stack, $element;
	if (%attr) {
	    foreach (keys %attr) {
		$xml .= " $_=\"" . $attr{$_} . "\"";
	    }
	}
	$xml .= ">";
	$oai->store_xml(\$xml);
    } elsif ($oai->{handles}->{$element}) {
	$oai->{triggered} = 1;
	$stack = $oai->{stack} = [];
	if ($debug) {
	    print "Resetting at tag '$element'\n";
	}
	$oai->reset_xml();
	handle_start($expat, $element, %attr);
    } elsif ($element eq $resumptionToken) {
	$oai->{incomplete} = 1;
	$oai->{cursor} = $attr{cursor};
	$oai->{completeListSize} = $attr{completeListSize};
	$oai->{expirationDate} = $attr{expirationDate};
    } else {
	if ($debug) {
	    print "Neglecting tag '$element'\n";
	}
    }
}


sub handle_end {
## Expat element end handler

    my ($expat, $element) = @_;
    my $oai = get_obj($expat);
    my $xml = "</$element>";
    my $stack = $oai->{stack};

    if ($oai->{triggered}) {
	unless ($element eq pop @$stack) {
	    print "Non mismatching begin-/end-tag in line " . $expat->current_line();
	} 
	$oai->store_xml(\$xml);
	unless (@$stack) {
	    if (my $callback = $oai->get_callback($element)) {
		if ($debug) {
		    print "Flushing\n";
		}
		&$callback($oai, "$xml_header\n".$oai->get_xml());
		$oai->{harvested} ++;
	    }
	    $oai->{triggered} = 0;
	}
    }
}


sub handle_char {
## Expat character handler

    my ($expat, $str) = @_;
    my $oai = get_obj($expat);
    
    $str =~ s/</&lt;/;
    $str =~ s/>/&gt;/;
    $str =~ s/&amp;/&/;
    $str =~ s/&/&amp;/;
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    if ($oai->{triggered}) {
	$oai->store_xml(\$str);
    } elsif ($oai->{incomplete} && !defined($oai->{$resumptionToken})) {
	$oai->{$resumptionToken} = $str;
    }
}


1;

## $Log: MyOAI.pm,v $
## Revision 1.5  2002/11/27 14:07:30  sondberg
## Nothing important.
##
## Revision 1.4  2002/11/01 12:31:40  sondberg
## Added some statistical book-keeping.
##
## Revision 1.3  2002/11/01 12:20:45  sondberg
## Implemented handling of resumptionToken.
##
## Revision 1.2  2002/10/31 09:39:08  sondberg
## Added some documentation.
##
## Revision 1.1  2002/10/31 09:18:35  sondberg
## Added this file to repository.
##
